//================================================================
// HW 3(a) Matrix multiply continued -- Moderately sized matrices
// 
// Spectral methods offer accuracy that increases exponentially with
// the number of grid-points.  To reap this benefit, one needs to
// choose the location of grid-points carefully.  Furthermore,
// operations like differentiation become dense (i.e. for a grid with
// N points, one would use all N points go compute derivatives).


#include <cmath>
#include <iostream>
#include <iomanip>



// return the correct location of the grid-points, given a grid with N
// points.  The pointer 'x' must point to an allocated array of length
// N+1
void ChebyshevCollocationPoints(const int N, double* x) {
  for(int k=0; k<=N; ++k)
    x[k] = cos(M_PI *k / N );
};


// differentiation matrix for a Chebyshev-grid with collocation points
// x.  The pointer 'x' must point to an allocated array of length N+1,
// filled with ChebyshevCollocationPoints().  The pointer result must
// point to an allocated array of length (N+1)*(N+1).
//
// result[(N+1)*i+j] = D_{ij} 

//  (code not optimized, because the matrix will be computed only
// once)
void DifferentiationMatrix(const int N, const double* x, double* D) {
  const int n=N+1; // translate from spectral notation to programming notation
  D[0*n+0] =  (2.*N*N+1)/6.;
  D[N*n+N] = -(2.*N*N+1)/6.;
  for(int j=1; j<N; ++j) 
    D[j*n+j] = -x[j]/(2*(1-x[j]*x[j]));

  // off-diagonal elements
  for(int i=0; i<=N; ++i) {
    for(int j=0; j<=N; ++j) {
      if(i!=j) {
	const double sgn = 1-2.*((i+j) % 2);  // (-1)^(i+j) 
	//std::cout << "i+j=" << i+j << ", sgn=" << sgn << std::endl;
	D[i*n+j]= sgn/(x[i]-x[j]); 
	// adjust for edge:
	if(i==0 || i==N) D[i*n+j] *= 2;
	if(j==0 || j==N) D[i*n+j] *= 0.5;
      }
    }
  }
}

//================================================================
// trial function 
//================================================================
double f(const double x, const double y) {
  //return x*x*x;
  return exp(sin(6*x))*cos(14*x*y);
}

void df(const double x, const double y,
	double& dfdx, double& dfdy) {
  dfdx=(6*cos(6*x)*cos(14*x*y)-14*y*sin(14*x*y))*exp(sin(6*x));
  dfdy=-14*x*sin(14*x*y)*exp(sin(6*x));
}


//================================================================
// Test routine:  Differentiate & compare to analytical answer
//================================================================

void TwoD_Test(const int Nx, const int Ny) {
  const int nx=Nx+1;
  const int ny=Ny+1;

  // setup of spectral information
  double* x=new double[nx];
  double* y=new double[ny];
  double* Dx=new double[nx*nx];  
  double* Dy=new double[nx*nx];  
  ChebyshevCollocationPoints(Nx,x);
  ChebyshevCollocationPoints(Ny,y);
  DifferentiationMatrix(Nx,x,Dx);
  DifferentiationMatrix(Ny,y,Dy);

  // vector to be differentiated
  double* data=new double[nx*ny];
  for(int i=0; i<nx; ++i) 
    for(int j=0; j<ny; ++j)
      data[i*ny+j] = f(x[i],y[j]);

  // storage for derivatives
  double* ddatadx=new double[nx*ny];
  double* ddatady=new double[nx*ny];

  //================
  // Compute x-Derivative
  //================
  // this, of course, is a very inefficient implementation.
  for(int i=0; i<nx; ++i) {
    for(int j=0; j<ny; ++j) {
      ddatadx[i*ny+j]=0;
      for(int k=0; k<nx; ++k) {
	ddatadx[i*ny+j] += Dx[i*nx+k]*data[k*ny+j];
      }
    }
  }
  
  //================
  // Compute y-Derivative
  //================
  // this, of course, is a very inefficient implementation.
  for(int i=0; i<nx; ++i) {
    for(int j=0; j<ny; ++j) {
      ddatady[i*ny+j]=0;
      for(int k=0; k<ny; ++k) {
	ddatady[i*ny+j] += Dy[j*ny+k]*data[i*ny+k];
      }
    }
  }


  //================
  // Compute error
  //================
  double errx=0, erry=0;
  for(int i=0; i<nx; ++i) {
    for(int j=0; j<ny; ++j) {
      double fx,fy;
      df(x[i], y[j], fx, fy);
      double tmp=ddatadx[i*ny+j]-fx;
      errx += tmp*tmp;
      tmp=ddatady[i*ny+j]-fy;
      erry += tmp*tmp;
    }
  }
  errx=sqrt(errx/(nx*ny));
  erry=sqrt(erry/(nx*ny));

  //================
  // Output error
  //================

  std::cout << Nx << "  " << errx << "  " << erry << std::endl;
}

int main() {
  std::cout << "Sizes relevant for the test function f" << std::endl;
  for(int N=8; N<=32; N+=8)
    TwoD_Test(2*N,N);

  std::cout << "Now for sheer size (irrespective of need)" << std::endl;
  for(int N=128; N<=1048; N*=2)
    TwoD_Test(2*N,N);
}
